body, td, th {
    font-family: Verdana;
    font-size: 75%;
}

table {
    border-collapse: collapse;
    border: 1px solid black;
}

td, th {
    border: 1px solid #cccccc;
    padding: 3px;
}

.toggle-indicator {
    background-color: darkgreen;
    color: white;
    font-weight:bold;
}

#topBar {
  position: fixed;
  top: 0;
  margin-left: 205px;                                   /* width of left bar */
  height: 60px;             /* width of top bar */
  width: 100%;
  color: black;
  font-size: 1em;
  background-color: #eeeeee;
  padding-left: 20px;
  padding-top: 5px;
}

#topBar button { font-size: 1em; }

#content {
    padding-top: 5px;
    padding-left: 20px;
    margin-top: 60px;               /* width of top bar */
    margin-left: 205px;                                      /* width of left bar */
    font-size: 1em;
}

#leftCol {
    position: fixed;
    width: 205px;                                            /* width of left bar */
    overflow-y: scroll;
    top: 0;
    bottom: 0;
    font-size: 1em;
    padding-right: 5px;
    padding-bottom: 30px;
    padding-top: 10px;
    /* word-break: break-all; */
}

a.anchor {
    display: block;
    position: relative;
    top: -60px;                 /* width of top bar */
    visibility: hidden;
}

/* Three image containers (use 25% for four, and 50% for two, etc) */
.column {
  float: left;
  width: 40%;
  padding: 5px;
}

/* Clear floats after image containers */
.row::after {
  content: "";
  clear: both;
  display: table;
}

img.map { 
  height: 400px;
}

img.chart {
  height: 350px; 
}

img.trend-chart {
  height: 200px;
}